# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..proxima.simple_index import build_index, search_index


METRIC_TO_PROXIMA_METRIC_TYPE = {
    'l2': 'Euclidean',
    'euclidean': 'Euclidean',
    'canberra': 'Canberra',
    'chebyshev': 'Chebyshev',
    'sqeuclidean': 'SquaredEuclidean',
    'innerproduct': 'InnerProduct',
    'manhattan': 'Manhattan'
}

build_proxima_index = build_index
proxima_query = search_index
